using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;

namespace MEdiaSpaceCore
{
    [Serializable]
    class VideoPacket
    {
        private int _width;
        private int _height;
        private Constants.VIDEO_PROTOCOLS _encoding;
        private byte[] _imageData;

        [NonSerialized]
        private Bitmap _decodedImageData = null;

        public VideoPacket(Bitmap bitmap, Constants.VIDEO_PROTOCOLS encoding) : this(bitmap, encoding, 50)
        {
        }

        public VideoPacket(Bitmap bitmap, Constants.VIDEO_PROTOCOLS encoding, int compressionRatio)
        {
            if (bitmap == null)
                throw new ArgumentNullException("bitmap");

            // Create an encoded representation of the captured image in a byte buffer
            BitmapData data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                                              ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            // Store the "header" data
            _width = bitmap.Width;
            _height = bitmap.Height;
            _encoding = encoding;

            switch (_encoding)
            {
                case Constants.VIDEO_PROTOCOLS.RAW:
                    int imageDataLength = bitmap.Width * bitmap.Height * Constants.BYTES_PER_PIXEL;

                    unsafe
                    {
                        _imageData = new byte[imageDataLength];
                        Marshal.Copy(data.Scan0, _imageData, 0, imageDataLength);
                    }
                    break;

                case Constants.VIDEO_PROTOCOLS.JPEG:
                    EasyImages.JPEG encoded = new EasyImages.JPEG(bitmap, compressionRatio);
                    _imageData = encoded.GetByteArray();
                    break;

                default:
                    throw new Exception("Invalid encoding protocol for video packet encoding");
            }

            bitmap.UnlockBits(data);
        }

        public int getWidth()
        {
            return _width;
        }

        public int getHeight()
        {
            return _height;
        }

        public Bitmap getBitmap()
        {
            if (_decodedImageData == null)
            {
                switch (_encoding)
                {
                    case Constants.VIDEO_PROTOCOLS.RAW:
                        unsafe
                        {
                            fixed (byte* imageDataAddr = _imageData)
                            {
                                IntPtr imageDataPtr = (IntPtr)imageDataAddr;
                                _decodedImageData = new Bitmap(_width, _height, _width * Constants.BYTES_PER_PIXEL,
                                                               PixelFormat.Format24bppRgb, imageDataPtr);
                            }
                        }
                        break;

                    case Constants.VIDEO_PROTOCOLS.JPEG:
                        MemoryStream memStream = new MemoryStream();
                        memStream.Write(_imageData, 0, _imageData.Length);
                        _decodedImageData = new Bitmap(memStream);
                        memStream.Close();
                        memStream.Dispose();
                        break;

                    default:
                        throw new Exception("Invalid encoding protocol for video packet decoding");
                }
            }

            return _decodedImageData;
        }
    }
}
