using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Net;
using System.Text;
using System.Threading;
using System.Timers;
using System.Windows.Forms;
using GT.Net;
using MEdiaSpaceCore;

namespace MEdiaSpaceOffice
{
    partial class OfficeForm : CustomForm
    {
        // A couple of constants to make the code easier to read
        private bool NOT_CONNECTED = false;
        
        // Private status variable
        protected bool _talking = false;

        // Shared variable manager
        Variables _variables = null;

        // Webcam source variable
        protected VideoWrapper _videoWrapper = null;

        // Audio wrapper variable
        protected AudioWrapper _audioWrapper = null;

        // Networking bits
        private ClientRepeater _server = null;
        private Client _client = null;
        private ISessionStream _sessionStream = null;
        private IObjectStream _videoStream = null;
        private IObjectStream _valueStream = null;
        private System.Timers.Timer _timer = null;

        // SharedPhidgets variables
        private GroupLab.SharedPhidgets.ConnectionManager _connectionManager = null;
        private GroupLab.SharedPhidgets.InterfaceKit _interfaceKit = null;

        // UI status variables
        private bool _isDragging = false;
        private int _dragX, _dragY;


        //
        // Lifecycle
        //
        public OfficeForm()
        {
            InitializeComponent();
        }


        //
        // UI change handlers
        //
        private void OfficeForm_Load(object sender, EventArgs e)
        {
            // Shared variable setup
            _variables = new Variables();
            _variables.DoorActuated += new Variables.DoorActuationHandler(_variables_DoorActuated);
            _variables.HomeMessageChanged += new Variables.MessageChangeHandler(_variables_HomeMessageChanged);
            _variables.HomeNoteTextChanged += new Variables.NoteTextChangeHandler(_variables_HomeNoteTextChanged);
            _variables.HomeSkypeHandleChanged += new Variables.SkypeHandleChangeHandler(_variables_HomeSkypeHandleChanged);
            _variables.HomeStatusChanged += new Variables.StatusChangeHandler(_variables_HomeStatusChanged);
            _variables.OfficeMessageChanged += new Variables.MessageChangeHandler(_variables_OfficeMessageChanged);
            _variables.OfficeVideoConfigurationChanged += new Variables.VideoConfigurationChangeHandler(_variables_OfficeVideoConfigurationChanged);

            // UI setup
            SetDocumentButtonEnabled(false);
            SetPhoneButtonEnabled(false);
            SetPhoneButtonStatus(NOT_CONNECTED);
            SetRemoteImage(null);

            // Networking setup
            _server = new ClientRepeater(Constants.OFFICE_PORT);
            _server.Start();

            _client = new Client(new MEdiaSpaceClientConfiguration());
            _client.ErrorEvent += new ErrorEventNotication(_client_ErrorEvent);
            _client.Start();
            ChannelDeliveryRequirements cdr = new ChannelDeliveryRequirements(Reliability.Reliable, MessageAggregation.Immediate, Ordering.Ordered);
            _sessionStream = _client.GetSessionStream(Constants.OFFICE_IP, Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.SESSION_CHANNEL, cdr);
            if (_sessionStream == null)
                throw new Exception("Problem monitoring session state of GT# library");
            _videoStream = _client.GetObjectStream(Constants.OFFICE_IP, Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.VIDEO_CHANNEL, cdr);
            _valueStream = _client.GetObjectStream(Constants.OFFICE_IP, Convert.ToString(Constants.OFFICE_PORT), (byte)Constants.DATA_CHANNEL, cdr);
            _variables.SetStream(_valueStream);

            // Video feed setup
            _videoWrapper = new VideoWrapper(_variables.officeProtocol,
                                             _variables.officeFrameRateSlow,
                                             _variables.officeCompression,
                                             new VideoWrapper.LocalImageSentHandler(SetLocalImage));
            _videoWrapper.ImageReceived += new VideoWrapper.RemoteImageReceivedHandler(SetRemoteImage);
            _videoWrapper.Start();

            // Audio feed setup
            _audioWrapper = new AudioWrapper(null);
            _variables.officeSkypeHandle = _audioWrapper.CurrentHandle();
            _audioWrapper.CallConnected += new AudioWrapper.CallConnectedHandler(audioWrapper_CallConnected);
            _audioWrapper.CallingAvailable += new AudioWrapper.CallingAvailableHandler(SetPhoneButtonEnabled);

            // SharedPhidgets setup
            if (components == null)
                components = new System.ComponentModel.Container();

            _connectionManager = new GroupLab.SharedPhidgets.ConnectionManager(components);
            _connectionManager.SharedDictionaryURL = "tcp://localhost:sp";
            _connectionManager.OpenConnection();

            _interfaceKit = new GroupLab.SharedPhidgets.InterfaceKit(components);
            _interfaceKit.AutoAttach = true;
            _interfaceKit.ComponentDescriptionPrefix = null;
            if (_variables.interfaceKitFirst == false)
                _interfaceKit.FilterSerialNumbers.Add(Convert.ToString(_variables.interfaceKitSerial));
            _interfaceKit.GUID = "59aed5ca-3bc6-431e-97cb-690c4fe66327";
            _interfaceKit.Invisible = false;
            _interfaceKit.OverrideDesignTimeLock = false;
            _interfaceKit.Attach += new EventHandler(_interfaceKit_Attach);
            _interfaceKit.Detach += new EventHandler(_interfaceKit_Detach);
            _interfaceKit.InputChange += new GroupLab.SharedPhidgets.BitStateChangeEventHandler(_interfaceKit_InputChange);
            _interfaceKit.SensorChangeAsynchronous += new GroupLab.SharedPhidgets.IndexSensorChangeEventHandler(interfaceKit_SensorChangeAsynchronous);

            // Start up the networking bits
            _timer = new System.Timers.Timer(Constants.NETWORK_CHECK_RATE);
            _timer.Elapsed += new ElapsedEventHandler(_timer_Elapsed);
            _timer.Start();
        }

        private void OfficeForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            // Close the audio connection (if there is one)
            if (_audioWrapper != null && _audioWrapper.IsTalking())
                _audioWrapper.StopCall();

            // Shut off the video
            if (_videoWrapper != null)
            {
                _videoWrapper.Stop();
                _videoWrapper.SetStream(null);
            }

            // Wipe our identity for the audio and shut off the network
            if (_client != null)
            {
                _variables.officeSkypeHandle = null;
                _client.Stop();
            }
        }

        private void outgoingPictureBox_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            _isDragging = false;
        }

        private void outgoingPictureBox_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (_isDragging)
            {
                int newTop = outgoingPictureBox.Top + (e.Y - _dragY);
                int newLeft = outgoingPictureBox.Left + (e.X - _dragX);

                newTop = Math.Max(newTop, incomingPictureBox.Top + Constants.PBOX_MARGIN);
                newTop = Math.Min(newTop, incomingPictureBox.Bottom - Constants.PBOX_MARGIN - outgoingPictureBox.Height);

                newLeft = Math.Max(newLeft, incomingPictureBox.Left + Constants.PBOX_MARGIN);
                newLeft = Math.Min(newLeft, incomingPictureBox.Right - Constants.PBOX_MARGIN - outgoingPictureBox.Width);

                outgoingPictureBox.Top = newTop;
                outgoingPictureBox.Left = newLeft;
            }
        }

        private void outgoingPictureBox_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            _isDragging = true;

            _dragX = e.X;
            _dragY = e.Y;
        }

        private void chatInputBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                _variables.officeMessage = chatInputBox.Text;
                chatInputBox.Text = "";
            }
        }
        
        private void documentButton_Click(object sender, EventArgs e)
        {
        }

        private void phoneButton_Click(object sender, EventArgs e)
        {
            if (!_talking && _audioWrapper.CurrentHandle() != null)
            {
                Thread audioSourceThread = new Thread(new ThreadStart(_audioWrapper.StartCall));
                audioSourceThread.IsBackground = true;
                audioSourceThread.Start();
            }
            else
                _audioWrapper.StopCall();
        }


        //
        // GTClient session handler
        //
        private void _timer_Elapsed(object sender, ElapsedEventArgs e)
        {
            // If the network has died, stop trying to read from it
            if (_client == null || _client.Active == false)
            {
                _timer.Stop();
                _timer.Dispose();
                _timer = null;
                return;
            }

            _client.Update();

            // Handle session messages first
            SessionMessage sm;
            while ((sm = _sessionStream.DequeueMessage(0)) != null)
                HandleIncomingSessionMessage(sm);

            // Handle video-related messages next
            Object packet;
            while ((packet = _videoStream.DequeueMessage(0)) != null)
                _videoWrapper.HandleIncomingMessage(packet);

            // Handle data messages last
            while ((packet = _valueStream.DequeueMessage(0)) != null)
                _variables.HandleIncomingMessage(packet);
        }

        private void HandleIncomingSessionMessage(SessionMessage message)
        {
            if (message.Action == SessionAction.Joined && message.ClientId != _sessionStream.UniqueIdentity)
            {
                AddChatMessage("MEdiaSpace", "The remote client has connected successfully.");

                _variables.Populate();

                _videoWrapper.SetStream(_videoStream);

                SetConnected(true);
                SetDocumentButtonEnabled(true);
                SetPhoneButtonEnabled(true);
                SetPhoneButtonStatus(NOT_CONNECTED);
                SetRemoteImage(null);
            }
            else if (message.Action == SessionAction.Left)
            {
                SetConnected(false);
                SetDocumentButtonEnabled(false);
                SetPhoneButtonEnabled(false);
                SetPhoneButtonStatus(NOT_CONNECTED);
                SetRemoteImage(null);
                ClearChatSpool();
                AddChatMessage("MEdia Space", "The remote client has disconnected");

                // Tell the video wrapper to stop trying to send packets
                _videoWrapper.SetStream(null);

                // Hang up the audio connection (if it'_server still playing)
                if (_audioWrapper != null && _audioWrapper.IsTalking())
                    _audioWrapper.StopCall();

                // Wipe out the old Skype info
                _variables.homeSkypeHandle = null;
            }
        }

        void _client_ErrorEvent(ErrorSummary summary)
        {
            throw new Exception("The method or operation is not implemented.");
        }


        //
        // Audio wrapper change handler
        //
        private void audioWrapper_CallConnected(bool talking)
        {
            _talking = talking;
            SetPhoneButtonStatus(_talking);

            if (_talking)
                _videoWrapper.SetFramerate(_variables.officeFrameRateFast);
            else
                _videoWrapper.SetFramerate(_variables.officeFrameRateSlow);
        }


        //
        // SharedPhidgets change handlers
        //
        private void _interfaceKit_Attach(object sender, EventArgs e)
        {
            _variables.doorActuationAvailable = true;
        }

        private void _interfaceKit_Detach(object sender, EventArgs e)
        {
            _variables.doorActuationAvailable = false;
        }

        private void _interfaceKit_InputChange(object sender, GroupLab.SharedPhidgets.BitStateChangeEventArgs e)
        {
            if (e.Index == _variables.doorbellPort)
                _variables.doorbellStatus = e.State;
        }

        private void interfaceKit_SensorChangeAsynchronous(object sender, GroupLab.SharedPhidgets.IndexSensorChangeEventArgs e)
        {
            if (e.Index == _variables.motionSensorPort)
            {
                // TEST
                //Console.WriteLine("motion detection changed: {0}", e);
                // END TEST
            }
        }


        //
        // Shared variables change handlers
        //
        private void _variables_DoorActuated()
        {
            if (_interfaceKit != null && _interfaceKit.Attached)
            {
                _interfaceKit.Outputs[_variables.doorMotorPort].State = true;
                Thread.Sleep((int)(_variables.doorMotorDelay * 1000));
                _interfaceKit.Outputs[_variables.doorMotorPort].State = false;
                _variables.doorActuationAvailable = true;
            }
        }

        private void _variables_HomeMessageChanged(String message)
        {
            if (message != null)
                AddChatMessage("Home", message);
        }

        private void _variables_HomeNoteTextChanged(String noteText)
        {
            // TODO
        }

        private void _variables_HomeSkypeHandleChanged(String skypeHandle)
        {
            _audioWrapper.SetSinkUsername(skypeHandle);

            if (skypeHandle != null)
                SetPhoneButtonEnabled(true);
            else
                SetPhoneButtonEnabled(false);
        }

        private void _variables_HomeStatusChanged(String status)
        {
            DisplayTextInControl(statusLabel, status);
        }

        private void _variables_OfficeMessageChanged(string message)
        {
            if (message != null)
                AddChatMessage("Office", message);
        }

        private void _variables_OfficeVideoConfigurationChanged(Constants.VIDEO_PROTOCOLS protocol, int compression, double frameRateSlow, double frameRateFast)
        {
            if (_videoWrapper != null)
            {
                _videoWrapper.SetVideoProtocol(protocol);
                _videoWrapper.SetJPEGCompression(compression);
                if (_talking)
                    _videoWrapper.SetFramerate(frameRateFast);
                else
                    _videoWrapper.SetFramerate(frameRateSlow);
            }
        }


        //
        // Content distribution functions
        //
        private void SetConnected(bool connected)
        {
            if (connected) {
                if (_variables.homeStatus != null)
                    DisplayTextInControl(statusLabel, _variables.homeStatus);
                else
                    DisplayTextInControl(statusLabel, Constants.HOME_CONNECTED_STRING);
            }
            else
                DisplayTextInControl(statusLabel, Constants.HOME_NOT_CONNECTED_STRING);
        }

        private void SetDocumentButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(documentButton, enabled);
        }

        private void SetPhoneButtonEnabled(bool enabled)
        {
            // Have to use delegate magic since this can be a cross-thread call
            EnableControl(phoneButton, enabled);
        }

        private void SetPhoneButtonStatus(bool talking)
        {
            if (talking)
                phoneButton.Image = global::MEdiaSpaceOffice.Properties.Resources.redphone_large;
            else
                phoneButton.Image = global::MEdiaSpaceOffice.Properties.Resources.greenphone_large;
        }

        private void SetRemoteImage(Image image)
        {
            DisplayImageInPictureBox(incomingPictureBox, image);
        }

        private void SetLocalImage(Image image)
        {
            if (null != image)
                image.RotateFlip(RotateFlipType.Rotate180FlipY);

            DisplayImageInPictureBox(outgoingPictureBox, image);
        }

        private void AddChatMessage(String sender, String message)
        {
            DisplayTextInControl(chatOutputBox, chatOutputBox.Text + sender + ": " + message + System.Environment.NewLine);
            FocusOnControl(chatInputBox);
        }

        private void ClearChatSpool()
        {
            DisplayTextInControl(chatOutputBox, "");
        }
   }
}